@@ -3,6 +3,7 @@
 
 # JabberBot: A simple jabber/xmpp bot framework
 # Copyright (c) 2007-2011 Thomas Perl <thp.io/about>
+# $Id: 40d08966900594ddefb8d3302290c2be097ab2b8 $
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,16 +19,27 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-"""A simple framework for creating Jabber/XMPP bots and services"""
+"""
+A framework for writing Jabber/XMPP bots and services
+
+The JabberBot framework allows you to easily write bots
+that use the XMPP protocol. You can create commands by
+decorating functions in your subclass or customize the
+bot's operation completely. MUCs are also supported.
+"""
 
 import os
 import re
 import sys
+import cgi
 
 try:
     import xmpp
 except ImportError:
-    print >>sys.stderr, 'You need to install xmpppy from http://xmpppy.sf.net/.'
+    print >> sys.stderr, """
+    You need to install xmpppy from http://xmpppy.sf.net/.
+    On Debian-based systems, install the python-xmpp package.
+    """
     sys.exit(-1)
 
 import time
@@ -35,10 +47,11 @@ import inspect
 import logging
 import traceback
 
+# Will be parsed by setup.py to determine package metadata
 __author__ = 'Thomas Perl <m@thp.io>'
-__version__ = '0.13'
+__version__ = '0.14'
 __website__ = 'http://thp.io/2007/python-jabberbot/'
-__license__ = 'GPLv3 or later'
+__license__ = 'GNU General Public License version 3 or later'
 
 def botcmd(*args, **kwargs):
     """Decorator for bot command functions"""
@@ -60,6 +73,10 @@ class JabberBot(object):
 
     MSG_AUTHORIZE_ME = 'Hey there. You are not yet on my roster. Authorize my request and I will do the same.'
     MSG_NOT_AUTHORIZED = 'You did not authorize my subscription request. Access denied.'
+    MSG_UNKNOWN_COMMAND = 'Unknown command: "%(command)s". Type "help" for available commands.'
+    MSG_HELP_TAIL = 'Type help <command name> to get more info about that specific command.'
+    MSG_HELP_UNDEFINED_COMMAND = 'That command is not defined.'
+    MSG_ERROR_OCCURRED = 'Sorry for your inconvenience. An unexpected error occurred.'
 
     PING_FREQUENCY = 0 # Set to the number of seconds, e.g. 60.
     PING_TIMEOUT = 2 # Seconds to wait for a response.
@@ -90,15 +107,17 @@ class JabberBot(object):
         self.__status = None
         self.__seen = {}
         self.__threads = {}
-        self.__lastping = None
+        self.__lastping = time.time()
         self.__privatedomain = privatedomain
         self.__acceptownmsgs = acceptownmsgs
 
+        self.custom_message_handler = None
+
         self.commands = {}
         for name, value in inspect.getmembers(self):
             if inspect.ismethod(value) and getattr(value, '_jabberbot_command', False):
                 name = getattr(value, '_jabberbot_command_name')
-                self.log.debug('Registered command: %s' % name)
+                self.log.info('Registered command: %s' % name)
                 self.commands[name] = value
 
         self.roster = None
@@ -130,25 +149,25 @@ class JabberBot(object):
 
 ################################
 
-    def connect( self):
+    def connect(self):
         if not self.conn:
             if self.__debug:
                 conn = xmpp.Client(self.jid.getDomain())
             else:
-                conn = xmpp.Client(self.jid.getDomain(), debug = [])
+                conn = xmpp.Client(self.jid.getDomain(), debug=[])
 
             conres = conn.connect()
             if not conres:
                 self.log.error('unable to connect to server %s.' % self.jid.getDomain())
                 return None
-            if conres<>'tls':
+            if conres != 'tls':
                 self.log.warning('unable to establish secure connection - TLS failed!')
 
             authres = conn.auth(self.jid.getNode(), self.__password, self.res)
             if not authres:
                 self.log.error('unable to authorize with server.')
                 return None
-            if authres<>'sasl':
+            if authres != 'sasl':
                 self.log.warning("unable to perform SASL auth os %s. Old authentication method used!" % self.jid.getDomain())
 
             conn.sendInitPresence()
@@ -163,14 +182,18 @@ class JabberBot(object):
 
         return self.conn
 
-    def join_room(self, room, username=None):
+    def join_room(self, room, username=None, password=None):
         """Join the specified multi-user chat room"""
+        NS_MUC = 'http://jabber.org/protocol/muc'
         if username is None:
             username = self.__username.split('@')[0]
         my_room_JID = '/'.join((room, username))
-        self.connect().send(xmpp.Presence(to=my_room_JID))
+        pres = xmpp.Presence(to=my_room_JID)
+        if password is not None:
+	    pres.setTag('x',namespace=NS_MUC).setTagData('password',password)
+        self.connect().send(pres)
 
-    def quit( self):
+    def quit(self):
         """Stop serving messages and exit.
 
         I find it is handy for development to run the
@@ -195,9 +218,9 @@ class JabberBot(object):
         NS_TUNE = 'http://jabber.org/protocol/tune'
         iq = xmpp.Iq(typ='set')
         iq.setFrom(self.jid)
-        iq.pubsub = iq.addChild('pubsub', namespace = xmpp.NS_PUBSUB)
-        iq.pubsub.publish = iq.pubsub.addChild('publish', attrs = { 'node' : NS_TUNE })
-        iq.pubsub.publish.item = iq.pubsub.publish.addChild('item', attrs= { 'id' : 'current' })
+        iq.pubsub = iq.addChild('pubsub', namespace=xmpp.NS_PUBSUB)
+        iq.pubsub.publish = iq.pubsub.addChild('publish', attrs={ 'node' : NS_TUNE })
+        iq.pubsub.publish.item = iq.pubsub.publish.addChild('item', attrs={ 'id' : 'current' })
         tune = iq.pubsub.publish.item.addChild('tune')
         tune.setNamespace(NS_TUNE)
 
@@ -220,7 +243,7 @@ class JabberBot(object):
             tune.addChild('uri').addData(song['uri'])
 
         if debug:
-            print 'Sending tune:', iq.__str__().encode('utf8')
+            self.log.info('Sending tune: %s' % iq.__str__().encode('utf8'))
         self.conn.send(iq)
 
     def send(self, user, text, in_reply_to=None, message_type='chat'):
@@ -239,7 +262,7 @@ class JabberBot(object):
 
     def send_simple_reply(self, mess, text, private=False):
         """Send a simple response to a message"""
-        self.send_message( self.build_reply(mess,text, private) )
+        self.send_message(self.build_reply(mess, text, private))
 
     def build_reply(self, mess, text=None, private=False):
         """Build a message for responding to another message.  Message is NOT sent"""
@@ -254,32 +277,44 @@ class JabberBot(object):
         return response
 
     def build_message(self, text):
-        """Builds an xhtml message without attributes."""
-        text_plain = re.sub(r'<[^>]+>', '', text)
-        message = xmpp.protocol.Message(body=text_plain)
+        """Builds an xhtml message without attributes. If input is not valid xhtml-im fallback to normal."""
+        message = None # fixes local variable 'message' referenced before assignment - Thanks to Aleksandr
+        text_plain = re.sub(r'<[^>]+>', '', text) # Try to determine if text has xhtml-tags - TODO needs improvement
         if text_plain != text:
-            html = xmpp.Node('html', {'xmlns': 'http://jabber.org/protocol/xhtml-im'})
+            message = xmpp.protocol.Message(body=text_plain) # Create body w stripped tags for reciptiens w/o xhtml-abilities - FIXME unescape &quot; etc.
+            html = xmpp.Node('html', {'xmlns': 'http://jabber.org/protocol/xhtml-im'}) # Start creating a xhtml body
             try:
                 html.addChild(node=xmpp.simplexml.XML2Node("<body xmlns='http://www.w3.org/1999/xhtml'>" + text.encode('utf-8') + "</body>"))
                 message.addChild(node=html)
             except Exception, e:
                 # Didn't work, incorrect markup or something.
-                # print >> sys.stderr, e, text
-                message = xmpp.protocol.Message(body=text_plain)
+                self.log.debug('An error while building a xhtml message. Fallback to normal messagebody')
+                message = None # Fallback - don't sanitize invalid input. User is responsible!
+        if message is None:
+            message = xmpp.protocol.Message(body=text) # Normal body
         return message
 
     def get_sender_username(self, mess):
-        """Extract the sender's user name from a message""" 
+        """Extract the sender's user name from a message"""
         type = mess.getType()
-        jid  = mess.getFrom()
+        jid = mess.getFrom()
         if type == "groupchat":
             username = jid.getResource()
         elif type == "chat":
-            username  = jid.getNode()
+            username = jid.getNode()
         else:
             username = ""
         return username
 
+    def get_full_jids(self, jid):
+        """Returns all full jids, which belong to a bare jid
+
+        Example: A bare jid is bob@jabber.org, with two clients connected, which
+        have the full jids bob@jabber.org/home and bob@jabber.org/work."""
+        for res in self.roster.getResources(jid):
+            full_jid = "%s/%s" % (jid,res)
+            yield full_jid
+
     def status_type_changed(self, jid, new_status_type):
         """Callback for tracking status types (available, away, offline, ...)"""
         self.log.debug('user %s changed status to %s' % (jid, new_status_type))
@@ -303,9 +338,17 @@ class JabberBot(object):
                 presence.getType(), presence.getShow(), \
                 presence.getStatus()
 
-        if self.jid.bareMatch(jid) and not self.__acceptownmsgs:
-            # Ignore our own presence messages
-            return
+        if self.jid.bareMatch(jid):
+            # update internal status
+            if type_ != self.OFFLINE:
+               self.__status = status
+               self.__show = show
+            else:
+               self.__status = ""
+               self.__show = self.OFFLINE
+            if not self.__acceptownmsgs:
+               # Ignore our own presence messages
+               return
 
         if type_ is None:
             # Keep track of status message and type changes
@@ -370,33 +413,33 @@ class JabberBot(object):
             self.send(jid, self.MSG_NOT_AUTHORIZED)
             self.roster.Unauthorize(jid)
 
-    def callback_message( self, conn, mess):
+    def callback_message(self, conn, mess):
         """Messages sent to the bot will arrive here. Command handling + routing is done in this function."""
         self.__lastping = time.time()
 
         # Prepare to handle either private chats or group chats
-        type     = mess.getType()
-        jid      = mess.getFrom()
-        props    = mess.getProperties()
-        text     = mess.getBody()
+        type = mess.getType()
+        jid = mess.getFrom()
+        props = mess.getProperties()
+        text = mess.getBody()
         username = self.get_sender_username(mess)
 
         if type not in ("groupchat", "chat"):
             self.log.debug("unhandled message type: %s" % type)
             return
 
+        # Ignore messages from before we joined
+        if xmpp.NS_DELAY in props: return
+
+        # Ignore messages from myself
+        if self.jid.bareMatch(jid): return
+        
         self.log.debug("*** props = %s" % props)
         self.log.debug("*** jid = %s" % jid)
         self.log.debug("*** username = %s" % username)
         self.log.debug("*** type = %s" % type)
         self.log.debug("*** text = %s" % text)
 
-        # Ignore messages from before we joined
-        if xmpp.NS_DELAY in props: return
-
-        # Ignore messages from myself
-        if username == self.__username: return
-
         # If a message format is not supported (eg. encrypted), txt will be None
         if not text: return
 
@@ -416,23 +459,39 @@ class JabberBot(object):
         cmd = command.lower()
         self.log.debug("*** cmd = %s" % cmd)
 
-        if self.commands.has_key(cmd):
+        if self.custom_message_handler is not None:
+            # Try the custom handler first. It can return None
+            # if you want JabberBot to fall back to the default.
+            reply = self.custom_message_handler(mess, text)
+
+            # If your custom_message_handler returns True, it
+            # is assumed that the custom_message_handler has
+            # taken care of processing the message, so we do
+            # not process the message any further here.
+            if reply == True:
+                return
+        else:
+            reply = None
+
+        if reply is None and self.commands.has_key(cmd):
             try:
                 reply = self.commands[cmd](mess, args)
             except Exception, e:
-                reply = traceback.format_exc(e)
-                self.log.exception('An error happened while processing a message ("%s") from %s: %s"' % (text, jid, reply))
+                self.log.exception('An error happened while processing a message ("%s") from %s: %s"' % (text, jid, traceback.format_exc(e)))
+                reply = self.MSG_ERROR_OCCURRED
         else:
             # In private chat, it's okay for the bot to always respond.
             # In group chat, the bot should silently ignore commands it
             # doesn't understand or aren't handled by unknown_command().
-            default_reply = 'Unknown command: "%s". Type "help" for available commands.<b>blubb!</b>' % cmd
-            if type == "groupchat": default_reply = None
-            reply = self.unknown_command( mess, cmd, args)
+            if type == 'groupchat':
+                default_reply = None
+            else:
+                default_reply = self.MSG_UNKNOWN_COMMAND % {'command': cmd}
+            reply = self.unknown_command(mess, cmd, args)
             if reply is None:
                 reply = default_reply
         if reply:
-            self.send_simple_reply(mess,reply)
+            self.send_simple_reply(mess, reply)
 
     def unknown_command(self, mess, cmd, args):
         """Default handler for unknown commands
@@ -465,7 +524,7 @@ class JabberBot(object):
 
     @botcmd
     def help(self, mess, args):
-        """Returns a help string listing available options.
+        """   Returns a help string listing available options.
 
         Automatically assigned to the "help" command."""
         if not args:
@@ -475,25 +534,22 @@ class JabberBot(object):
                 description = 'Available commands:'
 
             usage = '\n'.join(sorted([
-                '%s: %s' % (name, (command.__doc__.strip() or '(undocumented)').split('\n', 1)[0])
+                '%s: %s' % (name, (command.__doc__ or '(undocumented)').strip().split('\n', 1)[0])
                 for (name, command) in self.commands.iteritems() if name != 'help' and not command._jabberbot_hidden
             ]))
-            usage = usage + '\n\nType help <command name> to get more info about that specific command.'
+            usage = '\n\n'.join(filter(None, [usage, self.MSG_HELP_TAIL]))
         else:
             description = ''
             if args in self.commands:
-                usage = self.commands[args].__doc__.strip() or 'undocumented'
+                usage = (self.commands[args].__doc__ or 'undocumented').strip()
             else:
-                usage = 'That command is not defined.'
+                usage = self.MSG_HELP_UNDEFINED_COMMAND
 
-        top    = self.top_of_help_message()
+        top = self.top_of_help_message()
         bottom = self.bottom_of_help_message()
-        if top   : top    = "%s\n\n" % top
-        if bottom: bottom = "\n\n%s" % bottom
+        return '\n\n'.join(filter(None, [top, description, usage, bottom]))
 
-        return '%s%s\n\n%s%s' % ( top, description, usage, bottom )
-
-    def idle_proc( self):
+    def idle_proc(self):
         """This function will be called in the main loop."""
         self._idle_ping()
 
@@ -505,7 +561,7 @@ class JabberBot(object):
         if self.PING_FREQUENCY and time.time() - self.__lastping > self.PING_FREQUENCY:
             self.__lastping = time.time()
             #logging.debug('Pinging the server.')
-            ping = xmpp.Protocol('iq',typ='get',payload=[xmpp.Node('ping',attrs={'xmlns':'urn:xmpp:ping'})])
+            ping = xmpp.Protocol('iq', typ='get', payload=[xmpp.Node('ping', attrs={'xmlns':'urn:xmpp:ping'})])
             try:
                 res = self.conn.SendAndWaitForResponse(ping, self.PING_TIMEOUT)
                 #logging.debug('Got response: ' + str(res))
@@ -527,7 +583,7 @@ class JabberBot(object):
         """
         pass
 
-    def serve_forever( self, connect_callback = None, disconnect_callback = None):
+    def serve_forever(self, connect_callback=None, disconnect_callback=None):
         """Connects to the server and handles messages."""
         conn = self.connect()
         if conn:
@@ -553,4 +609,4 @@ class JabberBot(object):
         if disconnect_callback:
             disconnect_callback()
 
-
+# vim: expandtab tabstop=4 shiftwidth=4 softtabstop=4
